#pragma once
#include "Types.h"
#include <fstream>

class MeshLoader {
protected:
    std::vector<Node> nodes;
    std::vector<FiniteElement> finite_elements;
    std::vector<BoundaryFiniteElement> surfaces;

private:
    Node get_node(int ID) {
        return nodes.at(ID);
    }
    int last_nodes() {
        return nodes.rbegin()->ID;
    }

    double volume(int ID1, int ID2, int ID3, int ID4) { //   
        Node A = get_node(ID1);
        Node B = get_node(ID2);
        Node C = get_node(ID3);
        Node D = get_node(ID4);
        double A1 = (B.cd[0] - A.cd[0]) * ((C.cd[1] - A.cd[1]) * (D.cd[2] - A.cd[2]) - (C.cd[2] - A.cd[2]) * (D.cd[1] - A.cd[1]));
        double A2 = -(B.cd[1] - A.cd[1]) * ((C.cd[0] - A.cd[0]) * (D.cd[2] - A.cd[2]) - (D.cd[0] - A.cd[0]) * (C.cd[2] - A.cd[2]));
        double A3 = (B.cd[2] - A.cd[2]) * ((C.cd[0] - A.cd[0]) * (D.cd[1] - A.cd[1]) - (D.cd[0] - A.cd[0]) * (C.cd[1] - A.cd[1]));
        return (A1 + A2 + A3) / 6.0;
    }

public:
    virtual void loadmesh(const std::string&, const std::string&, const std::string&) = 0;
    virtual ~MeshLoader() = default;

    std::vector<Node>& Nodes() {
        return nodes;
    }
    std::vector<FiniteElement>& Finite_Elements() {
        return finite_elements;
    }
    std::vector<BoundaryFiniteElement>& Surfaces() {
        return surfaces;
    }
    void print() {
        std::cout << Nodes() << std::endl << std::endl
            << Finite_Elements() << std::endl << std::endl
            << Surfaces() << std::endl << std::endl;
    }

    std::vector<FiniteElement> finite_elements_ID(int, int, int);
    std::vector<FiniteElement> finite_elements_Edge(int, int);
    std::vector<Node> vertex_nodes_by_surface_ID(int);
    std::vector<FiniteElement> finite_elements_material(int);
    std::vector<BoundaryFiniteElement> surfaces_ID(int);
    void middle_insert();
    std::vector<std::vector<Node>> near_nodes();

    void volume_range(int, double, double); //   

};

